# SpringCloud GateWay 实现路由动态路由

 - [ ] **需求背景**
 
 在我们项目开发中，SpringCloud作为一款优秀的微服务架构和解决方案，使用率还是挺高的，其中使用GateWay来实现网关路由分发、负载均衡等功能，
 当然还可以利用GateWay来实现限流（比如RateLimter令牌桶算法等，也可搭载Sentinel来实现更细颗粒度的限流控制）、灰度发布、请求过滤处理等
 功能，由于这些不在本项目示例范围内，故不做相应的展开说明。而使用GateWay实现网关路由分发功能，一般通过两种方式来进行配置：
  1. 在GateWay的配置文件中配置路由规则（这是常用的方式），如下图所示：
 ![配置文件配置路由规则](images/gateway静态路由规则配置方法.png)
  2. 代码方式，通过代码方式在SpringContext中注入RouteLocator这个Bean，来实现路由规则的加载，如图所示：
 ![code方式配置路由规则](images/代码方式配置gateway路由规则.png)
 但是这两种方式只是在项目启动时初始化bean时加载RouteDefinition，若路由规则发生改变时候需要重启Gateway实例后方可重新加载新配置的路由规则。
 这就很难满足我们项目中的一些动态化的配置需求，尤其saas平台下多租户的应用和后端微服务实例为物理隔离的单独部署方式，灵活多变的部署方式带来网关路由
 规则的多样性和动态化，因此本人做了此gateway网关动态路由规则配置的项目示例。
 
 
  - [ ] **路由规则加载原理** 
  
 gateway网关程序在启动时，配置在配置文件或者代码中的路由信息会加载到内存中，路由信息被封装到RouteDefinition对象中，通过RouteDefinitionLocator接口来实现
 实现类来进行控制。配置文件中的配置信息是通过PropertiesRouteDefinitionLocator来进行加载。代码中的路由配置信息则是通过RouteLocatorBuilder构建出RouteLocator
 实例来加载。
 RouteDefinition主要由id、uri以及断言、过滤器等信息构成。
 RouteDefinitionRepository接口继承了RouteDefinitionLocator接口和RouteDefinitionWriter接口，故该接口具有RouteDefinition的查询、保存与删除等功能，默认使用InMemoryRouteDefinitionRepository类。
 而ApplicationEventPublisher实现RouteDefinition信息变动的发布与更新等操作，因此要实现路由规则动态加载（即热加载），只要我们使用RouteDefinitionWriter接口来实现RouteDefinition的增删功能，并使用ApplicationEventPublisher
 完成RouteDefinition信息变动的发布即可。
 
 
  - [ ] **动态路由实现方法*
  
  1.实现ApplicationEventPublisherAware接口，通过setApplicationEventPublisher方法获取ApplicationEventPublisher实例，并通过ApplicationEventPublisher
 实现RouteDefinition变更的事件发布操作。
 
  2.通过在我们自己定义的ApplicationEventPublisherAware实现类中注入RouteDefinitionWriter接口（RouteDefinitionRepository接口继承了RouteDefinitionWriter）,实现RouteDefinition的增删功能（修改功能也是通过增删功能来实现的）。
  
  3.对于单实例网关，通过本地操作可以实现本地路由信息的动态更新，但由于我们在项目中一般会根据项目情况部署多个网关实例，因此需要多个网关实例同步刷新配置信息方可。
  对此本人想了三种方案：
  1）在配置文件中配置所有的网关实例部署的地址（在nacos中配置也可），然后调用每个网关实例中暴露出来的路由信息刷新接口进行相应的操作，这种方法比较笨重，淘汰。
  2）通过消息队列的广播模式，每个网关实例作为消费端消费推送过来的路由信息进行本地更新操作。该方法由于需要借助第三方MQ，所以使用起来也不方便，淘汰。
  3）利用nacos的监听机制，每个网关实例监听到nacos server中配置信息的变动后进行本地化路由信息的刷新，由于nacos在我们yd等项目中已经作为服务发现和配置中心，因此此方案最佳
  也是本项目中所采纳的方案。

  - [ ] **动态路由实现步骤（重点）**
  
  1.路由规则动态配置架构如下：
  
   ![路由规则动态配置架构图](images/路由规则动态配置架构图.png)
   
  2.执行过程逻辑
  
  1）运维人员通过Nacos Server管理后台或我们自己平台管理后端（平台管理后端通过调用网关实例中RouteController暴露出来Rest API接口实现nacos server上路由动态配置信息的管理操作）来新增、修改、删除网关路由规则配置信息。
  2）在gateway上注册NacosConfigLisenter监听（在自定义NacosConfigListener类getConfig方法中），监听nacos Server上网关路由配置信息的变化情况。当配置信息发生变化时，会执行本地网关实例中路由配置信息的更新
  操作（在自定义NacosConfigListener类getConfig方法中调用的refreshLocalRoute方法进行操作）。
  3）在refreshLocalRoute方法中进行本地路由配置信息更新前，要先从内存中将已配置的路由信息删除完毕后再进行执行新增路由配置信息的操作。并在操作完成后
  将更新过的RouteDefinition的信息存入本地，以便后面出现新的更新情况下，先执行删除操作。
  
  3.程序实现步骤：
  
  1）引入模块如下：
  
   ![依赖类库](images/引入依赖类库.png)
   
  2）添加配置信息：其中dataId和group信息是我们在nacos config上的配置坐标。
  
   ![添加配置信息](images/添加配置信息.png)
   
  3）代码结构如下：
  
   ![醒目结构图](images/项目结构图.png)
   
  4）bean包下定义了三个类，其中GatewayRouteDefinition类是程序对外暴露的数据结构，GatewayFilterDefinition是GatewayRouteDefinition中
  过滤器的描述类，GatewayPredicateDefinition是GatewayRouteDefinition中断言的描述类，GatewayRouteDefinition除此外还包含id、uri、order等信息。
  
  5）service包下有两个类：
  #### DynamicRouteService   
  该类实现了ApplicationEventPublisherAware接口，在此类中setApplicationEventPublisher方法获得
  ApplicationEventPublisher接口的实例，并通过该实例可以操作路由配置信息的事件刷新。同时DynamicRouteService还注入了RouteDefinitionWriter接口。
  RouteDefinitionWriter接口被InMemoryRouteDefinitionRepository的接口类RouteDefinitionRepository所继承，因此可以在DynamicRouteService
  中使用RouteDefinitionWriter接口来实现路由规则配置信息的增删等功能。
  
  #### GatewayRoutesInNacosConfigService 
  该类主要封装了对Nacos Config的一些操作方法。主要提供给RouteController来进行调用，比较简单，不做过多介绍。
  
  6） util包下的GateWayNacosUtils类主要封装了获取ConfigService实例、从nacos server中获取route规则配置信息等操作。
  
  7）config包下的NacosConfigListener类
  #### GatewayRoutesInNacosConfigService 
  该类实现了CommandLineRunner接口，在网关实例启动后被执行，该类中最重要的就是getConfig方法，该方法实现了网关实例启动后初始化网关路由配置信息，并注册一个
  nacos监听，当nacos server上网关路由配置信息发生变更时，执行了refreshLocalRoute方法，该方法主要逻辑是首先要先删除网关实例内存中已经加载的配置信息，然后
  根据nacos server同步过来变更后的路由规则信息进行本地新增操作并发布加载到gateway工作内存中。
  
  8）controller包下的RouteController类主要封装了对nacos config server中网关路由信息的新增、查询、删除等操作的Rest API。
  
  9）路由定义规则配置信息说明
 #### yml中路由的配置信息
```spring:
   cloud:
     gateway:
       routes:
        - id: zz00001 # demo 中心路由
          uri: lb://zz00001 #应用服务名称可以使用店铺编码
          predicates:
           - Path=/mdxt/**
           - Header=orgCode,zz00001  #例如可根据店铺编码设置header 当客户端发request请求时携带orgCode=zz00001 则满足此规则，反之则不符合此路由规则
          filters:
            - RewritePath=/mdxt/(?<segment>.*),/$\{segment}

```
 ###  调用Rest API进行网关路由配置，数据结构如下
 POST: /route/addOrUpdate  请求体为json格式的requestBody
 ```
{
      "filters":[    
            {"args":{
            "_genkey_0": "/mdxt/(?<segment>.*)",
            "_genkey_1": "/${segment}"
          },
          "name":"RewritePath"
      }],
      "id":"zz00001",
      "order":0,
      "predicates":[
          {"args":{
            "_genkey_0":"/mdxt/**"
          },
          "name":"Path"
      },{"args":{
            "_genkey_0":"orgCode",
            "_genkey_1": "zz00001"
          },
           "name":"Header"
        }],
      "uri":"lb://zz00001"
     }
```
 ####  在nacos webui中添加和管理网关配置信息
 如果在nacos webui中添加网关配置信息，则需要用‘[ ]’将json数据体包起来，如下：
 ```
 [{
        "filters":[    
              {"args":{
              "_genkey_0": "/mdxt/(?<segment>.*)",
              "_genkey_1": "/${segment}"
            },
            "name":"RewritePath"
        }],
        "id":"zz00001",
        "order":0,
        "predicates":[
            {"args":{
              "_genkey_0":"/mdxt/**"
            },
            "name":"Path"
        },{"args":{
              "_genkey_0":"orgCode",
              "_genkey_1": "zz00001"
            },
             "name":"Header"
          }],
        "uri":"lb://zz00001"
       }]
```
 ###  测试demo模块中/demo/nativesql接口 通过网关访问http://localhost:9040/mdxt/demo/nativesql
 
  ![网关测试](images/postman通过gateway访问demo模块.png)
 